#!/usr/bin/env python
"""
    Export line detections and descriptors given a list of input images.
"""
import os
import argparse
import cv2
import numpy as np
import torch
import yaml
import os
import cairosvg
import iio

from model.line_detector import LineDetector

def load_config(config_path):
    """ Load configurations from a given yaml file. """
    # Check file exists
    if not os.path.exists(config_path):
        raise ValueError("[Error] The provided config path is not valid.")

    # Load the configuration
    with open(config_path, "r") as f:
        config = yaml.safe_load(f)

    return config

def export_descriptors(img_path, ckpt_path, config, device, extension,
                       out, multiscale=False):

    # Initialize the line matcher
    line_detector = LineDetector(
        config["model_cfg"], ckpt_path, device, config["line_detector_cfg"])
    print("\t Successfully initialized model")

    # Run the inference on each image and write the output on disk
    img = cv2.imread(img_path, 0)
    h, w = img.shape
    img = torch.tensor(img[None, None] / 255., dtype=torch.float,
                       device=device)

    # Run the line detection and description
    ref_detection = line_detector(img)
    ref_line_seg = ref_detection["line_segments"]
    ref_line_seg = ref_line_seg.reshape((-1, 4))

    ##### OUTPUT ######
    out_svg = out + '.svg'
    out_txt = out + '.txt'
    out_png = out + '.png'

    ftxt = open(out_txt, "w")
    fcsv = open(out_svg, "w")

    fcsv.write('<?xml version="1.0" standalone="no"?>\n<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN"\n"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd"><svg version="1.1" xmlns="http://www.w3.org/2000/svg"\nxmlns:xlink="http://www.w3.org/1999/xlink"\nwidth="'+str(w)+'" height="'+str(h)+'" viewBox="0 0 '+str(w)+' '+str(h)+'">')
    topng = '<?xml version="1.0" standalone="no"?>\n<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN"\n"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd"><svg version="1.1" xmlns="http://www.w3.org/2000/svg"\nxmlns:xlink="http://www.w3.org/1999/xlink"\nwidth="'+str(w)+'" height="'+str(h)+'" viewBox="0 0 '+str(w)+' '+str(h)+'">'

    color='black'
    width=1.5
    for tp_id, line in enumerate(ref_line_seg):
        y1, x1, y2, x2 = line # this is yxyx
        fcsv.write('<line x1="'+str(x1.item())+'" y1="' +str(y1.item())+ '" x2="' +str(x2.item())+ '" y2="' +str(y2.item())+ '" stroke="'+color+'" stroke-width="'+str(width)+'"/>\n')
        topng += '<line x1="'+str(x1.item())+'" y1="' +str(y1.item())+ '" x2="' +str(x2.item())+ '" y2="' +str(y2.item())+ '" stroke="white" stroke-width="'+str(width)+'"/>\n'
        ftxt.write(str(x1.item())+' '+str(y1.item())+' '+str(x2.item())+' '+str(y2.item())+'\n')
    fcsv.write('\n</svg>')
    topng += '\n</svg>'
    ftxt.close() 
    fcsv.close() 

    cairosvg.svg2png(bytestring=topng, write_to=out_png)
    im = iio.read(out_png)
    iio.write(out_png, im[:,:,:3])

if __name__ == "__main__":
    # Parse input arguments
    parser = argparse.ArgumentParser()
    parser.add_argument("--img_path", type=str, required=True,
                        help="path to the input image.")
    parser.add_argument("--out", type=str, required=True,
                        help="out struct.")
    parser.add_argument("--config", type=str,
                        default="config/export_line_features.yaml")
    parser.add_argument("--checkpoint_path", type=str,
                        default="pretrained_models/sold2_wireframe.tar")
    parser.add_argument("--multiscale", action="store_true", default=False)
    parser.add_argument("--extension", type=str, default=None)
    parser.add_argument("--detecThr", type=float, default=0.5)
    parser.add_argument("--inThr", type=float, default=0.99)
    args = parser.parse_args()

    device = torch.device("cpu")

    # Get the model config, extension and checkpoint path
    config = load_config(args.config)
    config['line_detector_cfg']["detect_thresh"] = args.detecThr
    config['line_detector_cfg']["inlier_thresh"] = args.inThr
    print(config)

    ckpt_path = os.path.abspath(args.checkpoint_path)
    extension = 'sold2' if args.extension is None else args.extension
    extension = "." + extension

    export_descriptors(args.img_path, ckpt_path, config, device, extension,
                       args.out, args.multiscale)
